/*
author: Sebastian Pelchat
date: october 2018
*/
#include "speedlimiter.hpp"

#define debug SpeedLimiter::pc->printf

Serial* SpeedLimiter::pc = new Serial(USBTX, USBRX);

SpeedLimiter::SpeedLimiter(const PinName& pedalInHi, const PinName& pedalInLo, const PinName& pedalOutHi, const PinName& pedalOutLo)
    : _pedalInHi(pedalInHi)
    , _pedalInLo(pedalInLo)
    , _pedalOutHi(pedalOutHi)
    , _pedalOutLo(pedalOutLo)
    , _referenceSpeed(DISABLE_ECO_ALGO_TRIGGER)
    , _measuredSpeed(0.0)
{
}

SpeedLimiter::~SpeedLimiter()
{
}

void SpeedLimiter::ipControllerTransferFunction()
{
//    static int counter = 0;
    
    float nextReferenceSpeed = getReferenceSpeed(true);
    setReferenceSpeed(nextReferenceSpeed);
//    pc->printf("Reference speed: %f\n\r", nextReferenceSpeed);
    
//    // une fois par 10 * 0.1 secondes -> 1 seconde
//    if (counter >= 10) {
//        counter = 0;
//        float nextReferenceSpeed = getReferenceSpeed(true);
//        setReferenceSpeed(nextReferenceSpeed);
//        pc->printf("Reference speed: %f\n\r", nextReferenceSpeed);
//    }
//    else { 
//        counter = counter + 1;
//    }
    // write voltages at beginning of function to prevent jitter
    // voltage will be delayed by 1 call which is okay.
//    pc->printf("H\n\r");
    const float voutHi = getOutputPedalVoltageHi();
    const float voutLo = getOutputPedalVoltageLo();

//    pc->printf("Hi: %f\t Lo: %f\n\r", voutHi, voutLo);

    writeAdcPedalHi(voutHi);
    writeAdcPedalLo(voutLo);

    // calculate voltage for next call
    const float referenceSpeed = getReferenceSpeed();
    float outputAdcVoltageHi = 0;
    float outputAdcVoltageLo = 0;

    if(referenceSpeed == DISABLE_ECO_ALGO_TRIGGER) {
        outputAdcVoltageHi = ecoDisabledAlgorithm();
    } else {
        outputAdcVoltageHi = ecoEnabledAlgorithm();
    }

//    outputAdcVoltageHi = ADC_OUTPUT_MAX_VALUE / 2;
    outputAdcVoltageLo = outputAdcVoltageHi / 2;

//    pc->printf("tmpHi: %f\t tmpLo: %f\n\r", outputAdcVoltageHi, outputAdcVoltageLo);

    setOutputPedalVoltageHi(outputAdcVoltageHi);
    setOutputPedalVoltageLo(outputAdcVoltageLo);
}

// Returns voltage read on analog input port chosen for pedal input 1
float SpeedLimiter::readAdcPedalHi()
{
    const float decPcValue = _pedalInHi.read();
    const float voltage = decPcValue * ADC_INPUT_MAX_VALUE;
    return voltage;
}

// Returns voltage read on analog input port chosen for pedal input 2
float SpeedLimiter::readAdcPedalLo()
{
    const float decPcValue = _pedalInLo.read();
    const float voltage = decPcValue * ADC_INPUT_MAX_VALUE;
    return voltage;
}

// Accepts a value in volts, converts to % and sets ADC for pedal output 1
void SpeedLimiter::writeAdcPedalHi(const float voltage)
{
    const float boundedValue = boundValue(voltage, PEDAL_HI_MIN_VALUE, PEDAL_HI_MAX_VALUE);
    const float decValue = voltageToDecimal(boundedValue, ADC_OUTPUT_MAX_VALUE);
    _pedalOutHi.write(decValue);
}

// Accepts a value in volts, converts to % and sets ADC for pedal output 2
void SpeedLimiter::writeAdcPedalLo(const float voltage)
{
    const float boundedValue = boundValue(voltage, PEDAL_LO_MIN_VALUE, PEDAL_LO_MAX_VALUE);
    const float decValue = voltageToDecimal(boundedValue, ADC_OUTPUT_MAX_VALUE);
    _pedalOutLo.write(decValue);
}

float SpeedLimiter::ecoDisabledAlgorithm()
{
    const float value = readAdcPedalHi();
    return value;
}

float SpeedLimiter::ecoEnabledAlgorithm()
{
    static bool first_acquisition = true;
    static time_t referenceTime = time(NULL); // seconds since January 1, 1970, a.k.a. since reset (time not set)
    
    // constants calibrated in vehicul
    double var = readAdcTest();
    double Kp =  18000 ;
    double Ki = 4300.0;

    // valeurs anterieures
    static double ie = 0.0;

    // calculs
    double Vm = getMeasuredSpeed();
    double Vd = getReferenceSpeed();
    //double Vd = 20;
    double out = 0.0;
    if (readAdcPedalHi() > 1) {
        double e = Vd - Vm;
        double eMax = 4;
        if(e > eMax) e = eMax;
        const double dt = TRANSFER_FUNCTION_PERIOD;
        ie = ie + e*dt;
        
        out = Kp*e + Ki*ie;

        // out est maintenant en 'force'
        out = (out / 133240.0 * (PEDAL_HI_MAX_VALUE-PEDAL_HI_MIN_VALUE)) + PEDAL_HI_MIN_VALUE;

        if(first_acquisition) {
            pc->printf("Acquisition start:\n\r");
            referenceTime = time(NULL);
            first_acquisition = false;
        }
    } else {
        ie = 0.00000000001;
        first_acquisition = true;
    }
    
    time_t elapsedTime = time(NULL) - referenceTime;
    pc->printf("%d: Vd: %.2f\tVm: %.2f\t\n\r", elapsedTime, Vd, Vm);

//    char timestamp[32];
//    strftime(timestamp, 32, "%I:%M %p\n", localtime(&seconds));
//    pc->printf("%s: Vd: %.2f\tVm: %.2f\t\n\r", timestamp, Vd, Vm);

    return (float)out;
}

// Returns 'value' bounded between 'lowerBound' and 'upperBound'
float SpeedLimiter::boundValue(float value, const float lowerBound, const float upperBound)
{
    if(value < lowerBound) {
        value = lowerBound;
    } else if(value > upperBound) {
        value = upperBound;
    }
    return value;
}

// Returns "value/reference" as a percentage in decimal form (0.5 for 50%)
float SpeedLimiter::voltageToDecimal(const float voltage, const float reference)
{
    return voltage/reference;
}
